package com.scbfkj.uni.service;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.jayway.jsonpath.JsonPath;
import com.scbfkj.uni.library.DataAliasGetUtil;
import com.scbfkj.uni.library.DataFormatUtil;
import com.scbfkj.uni.library.UniReturnUtil;
import com.scbfkj.uni.library.script.JavaScriptEngineUtil;
import com.scbfkj.uni.library.script.JsScriptEngineUtil;
import com.scbfkj.uni.process.DataBase;
import com.scbfkj.uni.system.Config;

import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.*;


public class DataProcessService {


    public static Map<String, Object> process(Map<String, Object> inData) {
        String lifecycleid = null;
        Map<String, Object> result;
        List<Map<String, Object>> resource = new ArrayList<>();
        resource.add(inData);
        String serviceId = null;
        Object algorithmlibraryid = null;
        Optional<String> serviceIdOpt = null;
        LocalDateTime startDateTime = LocalDateTime.now();
        try {
            serviceIdOpt = DataAliasGetUtil.getValue("serviceid", inData);
        } catch (Exception e) {
            LoggerService.logService(
                    startDateTime,
                    serviceId,
                    DataFormatUtil.toString(resource),
                    null,
                    LocalDateTime.now(),
                    null,
                    "-1",
                    e.getMessage(),
                    lifecycleid
            );
            return UniReturnUtil.fail(e.getMessage());
        }
        if (Objects.isNull(inData) || inData.isEmpty() || serviceIdOpt.isEmpty())
            return UniReturnUtil.fail("服务编号不能为空");
        serviceId = serviceIdOpt.get();

//        熔断

        List<Map<String, Object>> algorithmLibraries;
        try {
            if (serviceId.matches("^\\d+$")) {
                algorithmLibraries = DataBase.query(Config.centerConnectionStr, "select * from algorithmlibrary where serviceid=?", Collections.singletonList(new Object[]{serviceId}));
            } else {
                algorithmLibraries = DataBase.query(Config.localCenterConnectionStr, "select * from algorithmlibrary where serviceid=?", Collections.singletonList(new Object[]{serviceId}));
            }
        } catch (Exception e) {
//            记录异常
            LoggerService.logService(
                    startDateTime,
                    serviceId,
                    DataFormatUtil.toString(resource),
                    null,
                    LocalDateTime.now(),
                    null,
                    "-1",
                    e.getMessage(),
                    lifecycleid
            );
            return UniReturnUtil.fail(e.getMessage());
        }

        Optional<String> lifecycleidOpt = null;
        try {
            lifecycleidOpt = DataAliasGetUtil.getValue("lifecycleid", inData);
        } catch (Exception e) {
            LoggerService.logService(
                    startDateTime,
                    serviceId,
                    DataFormatUtil.toString(resource),
                    null,
                    LocalDateTime.now(),
                    null,
                    "-1",
                    e.getMessage(),
                    lifecycleid
            );
            return UniReturnUtil.fail(e.getMessage());
        }
        if (lifecycleidOpt.isEmpty()) {
            lifecycleid = DataAliasGetUtil.createLifeCycleCol(Config.containerCode, serviceId);
            inData.put("lifecycleid", lifecycleid);
        } else {
            lifecycleid = lifecycleidOpt.get();
        }

        for (Map<String, Object> algorithmLibrary : algorithmLibraries) {

            Map<String, Object> algorithmResult = null;

            long execTime;
            String code = null;
            String message = null;

            startDateTime = LocalDateTime.now();
            long startTime = System.currentTimeMillis();
            List<Object> parameters = null;
            algorithmlibraryid = algorithmLibrary.get("algorithmlibraryid");
            Object preConditions = algorithmLibrary.get("preconditions");
            Object preparameterset = algorithmLibrary.get("preparameterset");
            try {
                HashMap<String, Object> data = new HashMap<>();
                resource.add(data);
//                记录生命周期ID
                data.put("lifecycleid", lifecycleid);
                if (Objects.nonNull(preConditions)) {

                    String preCode;
                    if (Objects.isNull(preparameterset)) {
                        preCode = JsScriptEngineUtil.eval(DataFormatUtil.toString(preConditions), resource);
                    } else {
                        List<Object> params = getParams(Objects.requireNonNull(DataFormatUtil.toString(preparameterset)), resource);
                        preCode = JsScriptEngineUtil.eval(DataFormatUtil.toString(preConditions), params.toArray());
                    }
//                    记录前置条件结果
                    data.put("preResult", preCode);
                    if (Objects.equals("2", preCode)) {
                        LoggerService.logService(
                                startDateTime,
                                serviceId,
                                DataFormatUtil.toString(resource),
                                DataFormatUtil.toString(parameters),
                                LocalDateTime.now(),
                                DataFormatUtil.toString(algorithmResult),
                                "0",
                                message,
                                lifecycleid
                        );
                        return UniReturnUtil.success(resource.size() - 1);
                    } else if (Objects.equals("1", preCode)) {
                        continue;
                    }
                }

                Object parameterSet = algorithmLibrary.get("parameterset");
                Object dataSourceId = algorithmLibrary.get("datasourceid");
                List<Map<String, Object>> datasourceList = DataBase.query(Config.centerConnectionStr, "select * from datasource where datasourceid=?", Collections.singletonList(new Object[]{dataSourceId}));
                Map<String, Object> datasource = datasourceList.get(0);

//        获取入参列表
                parameters = new ArrayList<>();
                if (Objects.nonNull(parameterSet)) {
                    HashMap<String, Object> source = new HashMap<>();
                    source.put("args", resource);
                    source.put("algorithm", algorithmLibrary);
                    source.put("datasource", datasource);
                    parameters.addAll(getParams(parameterSet.toString(), source));
//                    算法入参
                    data.put("parameters", parameters);

                }
                List<String> filterColumns = Optional.of(resource).map(it -> it.get(0)).map(it -> it.get("filterColumns")).map(it -> ((List<String>) it)).orElse(new ArrayList<>());
                List<Map<String, Object>> filterLines = Optional.of(resource).map(it -> it.get(0)).map(it -> it.get("filterLines")).map(it -> ((List<Map<String, Object>>) it)).orElse(new ArrayList<>());

                algorithmResult = processByAlgorithm(algorithmLibrary, datasource, parameters, filterColumns, filterLines);
//                    算法执行结果
                data.put("result", algorithmResult);
                execTime = System.currentTimeMillis() - startTime;
//                    执行时长
                data.put("execTime", execTime);
//                    执行成功与否
                code = Optional.ofNullable(algorithmResult.get("code")).map(DataFormatUtil::toString).orElse(null);
                message = Optional.ofNullable(algorithmResult.get("message")).map(DataFormatUtil::toString).orElse(null);
                data.put("resultCode", code);
            } catch (Exception e) {
                message = e.getMessage();
//                    把错误细化后往上一层抛出统一记录日志
                throw new RuntimeException("算法id: %s 异常信息：%s ".formatted(algorithmlibraryid, e.getMessage()));
            } finally {
//                    不管成功还是失败都记录日志
                LoggerService.logService(
                        startDateTime,
                        serviceId,
                        DataFormatUtil.toString(resource),
                        DataFormatUtil.toString(parameters),
                        LocalDateTime.now(),
                        DataFormatUtil.toString(algorithmResult),
                        Optional.ofNullable(code).orElse("-1"),
                        message,
                        lifecycleid
                );
            }
            if (!Objects.equals(code, "0")) {
                return resource.get(resource.size() - 1);
            }
        }
        return resource.get(resource.size() - 1);

    }

    /**
     * @param algorithmLibrary 算法配置
     * @param datasource
     * @param parameters       算法参数
     * @param filterColumns
     * @param filterLines
     * @return
     * @throws Exception
     */
    public static Map<String, Object> processByAlgorithm(Map<String, Object> algorithmLibrary, Map<String, Object> datasource, List<Object> parameters, List<String> filterColumns, List<Map<String, Object>> filterLines) throws Exception {

        Object type = algorithmLibrary.get("algorithmtype");
        Object expression = algorithmLibrary.get("computingexpression");


        switch (type.toString()) {
//            java反射
            case "1" -> {
                Object invoke = JavaScriptEngineUtil.invoke(datasource, expression.toString(), parameters);
                return UniReturnUtil.success(invoke);
            }
//            JS表达式
            case "2" -> {
                return UniReturnUtil.success(JsScriptEngineUtil.eval(expression.toString(), parameters));
            }
//            数据库
            case "3" -> {
//                下放到Database中处理数据
//                参数表达式顺序是 数据源连接字符串（String.$.datasource.connectset）,sql表达式(String.$.algorithm.computingexpression)，需要操作的值(List.$.args[1].returnData)，执行编号(String.$.algorithm.executionnumber)
                return DataBase.exec(parameters.get(0).toString(), parameters.get(1).toString(), ((List<Map<String, Object>>) parameters.get(2)), parameters.get(3), filterColumns, filterLines);
            }
        }
        return UniReturnUtil.fail("算法类型不支持");
    }

    /**
     * @param script 前置表达式 不能为空
     * @param args   表达式参数
     * @return
     * @throws Exception
     */
    private static String preConditionScript(String script, List<Map<String, Object>> args, String parameterSet) throws Exception {
//        todo 兼容之前版本的写法的
//        判断算法中是否存在书名号
        if (Objects.isNull(parameterSet)) {
            return JsScriptEngineUtil.eval(script, args);
        } else {
            List<Object> params = getParams(parameterSet, args);
            return JsScriptEngineUtil.eval(script, params.toArray());
        }
    }

    /**
     * @param parameterSet 使用非标准的jsonPath表达式 String.$.a.b[0].c,String.abcc 多个参数之间使用;;分隔
     * @param source       jsonpath需要取值的数据源
     * @return
     */
    public static List<Object> getParams(String parameterSet, Object source) {
        String[] paths = parameterSet.split(";;");
        String json = DataFormatUtil.toString(source);
        List<Object> result = new ArrayList<>();

        for (String it : paths) {
            it = it.trim();
            String type = it.split("\\.")[0];
            String path = it.replace(type + ".", "");
            Object value = null;
//            $开头的使用标准的jsonPath
            if (path.startsWith("$")) {
                value = JsonPath.read(json, path);
            } else if (path.startsWith("#")) {
//                todo  缓存之类的特殊值

//                不是使用$#开头的直接取后面的字符串
            } else {
                value = path;
            }
            if (Objects.isNull(value)) {
                result.add(null);
                continue;
            }
            if ("String".equalsIgnoreCase(type)) {
                value = DataFormatUtil.toString(value);
            } else if ("List".equalsIgnoreCase(type)) {
                value = DataFormatUtil.toList(value);
            } else if ("Array".equalsIgnoreCase(type)) {
                value = DataFormatUtil.toArray(value);
            } else if ("Map".equalsIgnoreCase(type)) {
                value = DataFormatUtil.toMap(value);
            } else if ("Long".equalsIgnoreCase(type)) {
                value = Long.parseLong(value.toString());
            } else if ("Integer".equalsIgnoreCase(type)) {
                value = Integer.parseInt(value.toString());
            } else if ("Double".equalsIgnoreCase(type)) {
                value = Double.parseDouble(value.toString());
            } else if ("Float".equalsIgnoreCase(type)) {
                value = Float.parseFloat(value.toString());
            } else if ("Boolean".equalsIgnoreCase(type)) {
                value = Boolean.parseBoolean(value.toString());
            } else if ("Datetime".equalsIgnoreCase(type)) {
                String string = value.toString();
                String patten = "yyyy-MM-dd HH:mm:ss";
                if (string.contains("(")) {
                    patten = string.substring(string.indexOf("(") + 1, string.length() - 1);
                    string = string.substring(0, string.indexOf("("));
                }
                value = LocalDateTime.parse(string, DateTimeFormatter.ofPattern(patten));
            }

            result.add(value);
        }
        return result;
    }


}
